"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const anchor_1 = require("@coral-xyz/anchor");
const web3_js_1 = require("@solana/web3.js");
const console_table_printer_1 = require("console-table-printer");
const cloneDeep_1 = __importDefault(require("lodash/cloneDeep"));
const node_fetch_1 = __importDefault(require("node-fetch"));
const mangoAccount_1 = require("../src/accounts/mangoAccount");
const client_1 = require("../src/client");
const constants_1 = require("../src/constants");
const I80F48_1 = require("../src/numbers/I80F48");
const utils_1 = require("../src/utils");
const { MB_CLUSTER_URL } = process.env;
const GROUP_PK = '78b8f4cGCwmZ9ysPFMWLaLTkkaYnUjwMJYStWe5RTSSX';
async function buildClient() {
    const clientKeypair = new web3_js_1.Keypair();
    const options = anchor_1.AnchorProvider.defaultOptions();
    const connection = new web3_js_1.Connection(MB_CLUSTER_URL, options);
    const clientWallet = new anchor_1.Wallet(clientKeypair);
    const clientProvider = new anchor_1.AnchorProvider(connection, clientWallet, options);
    return await client_1.MangoClient.connect(clientProvider, 'mainnet-beta', constants_1.MANGO_V4_ID['mainnet-beta'], {
        idsSource: 'get-program-accounts',
    });
}
async function computePriceImpactOnJup(amount, inputMint, outputMint) {
    const url = `https://quote-api.jup.ag/v4/quote?inputMint=${inputMint}&outputMint=${outputMint}&amount=${amount}&swapMode=ExactIn&slippageBps=10000&onlyDirectRoutes=false&asLegacyTransaction=false`;
    const response = await (0, node_fetch_1.default)(url);
    try {
        let res = await response.json();
        res = res.data[0];
        return {
            outAmount: parseFloat(res.outAmount),
            priceImpactPct: parseFloat(res.priceImpactPct),
        };
    }
    catch (e) {
        console.log(url);
        console.log(e);
        throw e;
    }
}
async function computePriceImpactForLiqor(group, mangoAccounts, healthThresh, title) {
    // Filter mango accounts below a certain health ration threshold
    const mangoAccountsWithHealth = mangoAccounts
        .map((a) => {
        return {
            account: a,
            health: a.getHealth(group, mangoAccount_1.HealthType.liquidationEnd),
            healthRatio: a.getHealthRatioUi(group, mangoAccount_1.HealthType.liquidationEnd),
            liabs: (0, utils_1.toUiDecimalsForQuote)(a.getLiabsValue(group, mangoAccount_1.HealthType.liquidationEnd)),
        };
    })
        .filter((a) => a.healthRatio < healthThresh);
    const table = new console_table_printer_1.Table({
        columns: [
            { name: 'Coin', alignment: 'right' },
            { name: 'Oracle Price', alignment: 'right' },
            { name: 'On-Chain Price', alignment: 'right' },
            { name: 'Future Price', alignment: 'right' },
            { name: 'V4 Liq Fee', alignment: 'right' },
            { name: 'Liabs', alignment: 'right' },
            { name: 'Liabs slippage', alignment: 'right' },
            { name: 'Assets Sum', alignment: 'right' },
            { name: 'Assets Slippage', alignment: 'right' },
        ],
    });
    const USDC_MINT = 'EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v';
    const usdcBank = group.getFirstBankByMint(new web3_js_1.PublicKey(USDC_MINT));
    // For each token
    for (const banks of group.banksMapByMint.values()) {
        const bank = banks[0];
        const onChainPrice = (await (await (0, node_fetch_1.default)(`https://price.jup.ag/v4/price?ids=${bank.mint}`)).json())['data'][bank.mint.toBase58()]['price'];
        if (bank.tokenIndex === usdcBank.tokenIndex) {
            continue;
        }
        // Sum of all liabs, these liabs would be acquired by liqor,
        // who would immediately want to reduce them to 0
        // Assuming liabs need to be bought using USDC
        const liabs = 
        // Max liab of a particular token that would be liquidated to bring health above 0
        mangoAccountsWithHealth.reduce((sum, a) => {
            // How much would health increase for every unit liab moved to liqor
            // liabprice * (liabweight - (1+fee)*assetweight)
            const tokenLiabHealthContrib = bank.price.mul(bank.initLiabWeight.sub((0, I80F48_1.ONE_I80F48)().add(bank.liquidationFee).mul(usdcBank.initAssetWeight)));
            // Abs liab/borrow
            const maxTokenLiab = a.account
                .getEffectiveTokenBalance(group, bank)
                .min((0, I80F48_1.ZERO_I80F48)())
                .abs();
            // Health under 0
            const maxLiab = a.health
                .min((0, I80F48_1.ZERO_I80F48)())
                .abs()
                .div(tokenLiabHealthContrib)
                .min(maxTokenLiab);
            return sum.add(maxLiab);
        }, (0, I80F48_1.ZERO_I80F48)());
        const liabsInUsdc = 
        // convert to usdc, this is an approximation
        liabs
            .mul(bank.price)
            .floor()
            // jup oddity
            .min(I80F48_1.I80F48.fromNumber(99999999999));
        const pi1 = !liabsInUsdc.eq((0, I80F48_1.ZERO_I80F48)())
            ? await computePriceImpactOnJup(liabsInUsdc.toString(), USDC_MINT, bank.mint.toBase58())
            : { priceImpactPct: 0, outAmount: 0 };
        // Sum of all assets which would be acquired in exchange for also acquiring
        // liabs by the liqor, who would immediately want to reduce to 0
        // Assuming assets need to be sold to USDC
        const assets = mangoAccountsWithHealth.reduce((sum, a) => {
            // How much would health increase for every unit liab moved to liqor
            // assetprice * (liabweight/(1+liabliqfee) - assetweight)
            const liabBank = Array.from(group.banksMapByTokenIndex.values())
                .flat()
                .reduce((prev, curr) => prev.initLiabWeight.lt(curr.initLiabWeight) ? prev : curr);
            const tokenAssetHealthContrib = bank.price.mul(liabBank.initLiabWeight
                .div((0, I80F48_1.ONE_I80F48)().add(liabBank.liquidationFee))
                .sub(bank.initAssetWeight));
            // Abs collateral/asset
            const maxTokenHealthAsset = a.account
                .getEffectiveTokenBalance(group, bank)
                .max((0, I80F48_1.ZERO_I80F48)());
            const maxAsset = a.health
                .min((0, I80F48_1.ZERO_I80F48)())
                .abs()
                .div(tokenAssetHealthContrib)
                .min(maxTokenHealthAsset);
            return sum.add(maxAsset);
        }, (0, I80F48_1.ZERO_I80F48)());
        const pi2 = !assets.eq((0, I80F48_1.ZERO_I80F48)())
            ? await computePriceImpactOnJup(assets.floor().toString(), bank.mint.toBase58(), USDC_MINT)
            : { priceImpactPct: 0 };
        table.addRow({
            Coin: bank.name,
            'Oracle Price': bank['oldUiPrice'] < 0.1
                ? bank['oldUiPrice']
                : bank['oldUiPrice'].toFixed(2),
            'On-Chain Price': onChainPrice < 0.1 ? onChainPrice : onChainPrice.toFixed(2),
            'Future Price': bank._uiPrice < 0.1 ? bank._uiPrice : bank._uiPrice.toFixed(2),
            'V4 Liq Fee': (bank.liquidationFee.toNumber() * 100).toFixed(2) + '%',
            Liabs: (0, utils_1.toUiDecimalsForQuote)(liabsInUsdc).toLocaleString() + '$',
            'Liabs slippage': (pi1.priceImpactPct * 100).toFixed(2) + '%',
            'Assets Sum': ((0, utils_1.toUiDecimals)(assets, bank.mintDecimals) * bank.uiPrice).toLocaleString() + '$',
            'Assets Slippage': (pi2.priceImpactPct * 100).toFixed(2) + '%',
        });
    }
    const msg = title + '\n```\n' + table.render() + '\n```';
    console.log(msg);
    console.log();
}
async function computePerpPositionsToBeLiquidated(group, mangoAccounts, healthThresh, title) {
    const mangoAccountsWithHealth = mangoAccounts
        .map((a) => {
        return {
            account: a,
            health: a.getHealth(group, mangoAccount_1.HealthType.liquidationEnd),
            healthRatio: a.getHealthRatioUi(group, mangoAccount_1.HealthType.liquidationEnd),
            liabs: (0, utils_1.toUiDecimalsForQuote)(a.getLiabsValue(group, mangoAccount_1.HealthType.liquidationEnd)),
        };
    })
        .filter((a) => a.healthRatio < healthThresh);
    const table = new console_table_printer_1.Table({
        columns: [
            { name: 'Market', alignment: 'right' },
            { name: 'Price', alignment: 'right' },
            { name: 'Future Price', alignment: 'right' },
            { name: 'Notional Position', alignment: 'right' },
        ],
    });
    for (const pm of Array.from(group.perpMarketsMapByMarketIndex.values()).filter((pm) => !pm.name.includes('OLD'))) {
        const baseLots = mangoAccountsWithHealth
            .filter((a) => a.account.getPerpPosition(pm.perpMarketIndex))
            .reduce((sum, a) => {
            const baseLots = a.account.getPerpPosition(pm.perpMarketIndex).basePositionLots;
            const unweightedHealthPerLot = baseLots.gt(new anchor_1.BN(0))
                ? I80F48_1.I80F48.fromNumber(-1)
                    .mul(pm.price)
                    .mul(I80F48_1.I80F48.fromU64(pm.baseLotSize))
                    .mul(pm.initBaseAssetWeight)
                    .add(I80F48_1.I80F48.fromU64(pm.baseLotSize)
                    .mul(pm.price)
                    .mul((0, I80F48_1.ONE_I80F48)() // quoteInitAssetWeight
                    .mul((0, I80F48_1.ONE_I80F48)().sub(pm.baseLiquidationFee))))
                : pm.price
                    .mul(I80F48_1.I80F48.fromU64(pm.baseLotSize))
                    .mul(pm.initBaseLiabWeight)
                    .sub(I80F48_1.I80F48.fromU64(pm.baseLotSize)
                    .mul(pm.price)
                    .mul((0, I80F48_1.ONE_I80F48)()) // quoteInitLiabWeight
                    .mul((0, I80F48_1.ONE_I80F48)().add(pm.baseLiquidationFee)));
            const maxBaseLots = a.health
                .min((0, I80F48_1.ZERO_I80F48)())
                .abs()
                .div(unweightedHealthPerLot.abs())
                .min(I80F48_1.I80F48.fromU64(baseLots).abs());
            return sum.add(maxBaseLots);
        }, (0, I80F48_1.ONE_I80F48)());
        const notionalPositionUi = (0, utils_1.toUiDecimalsForQuote)(baseLots.mul(I80F48_1.I80F48.fromU64(pm.baseLotSize).mul(pm.price)));
        table.addRow({
            Market: pm.name,
            Price: pm['oldUiPrice'] < 0.1 ? pm['oldUiPrice'] : pm['oldUiPrice'].toFixed(2),
            'Future Price': pm._uiPrice < 0.1 ? pm._uiPrice : pm._uiPrice.toFixed(2),
            'Notional Position': notionalPositionUi.toLocaleString() + '$',
        });
    }
    const msg = title + '\n```\n' + table.render() + '\n```';
    console.log(msg);
    console.log();
}
async function logLiqorEquity(client, group, mangoAccounts, title) {
    const table = new console_table_printer_1.Table({
        columns: [
            { name: 'Account', alignment: 'right' },
            { name: 'Equity', alignment: 'right' },
        ],
    });
    // Filter mango accounts which might be closed
    const liqors = (await client.connection.getMultipleAccountsInfo(mangoAccounts))
        .map((ai, i) => {
        return { ai: ai, pk: mangoAccounts[i] };
    })
        .filter((val) => val.ai)
        .map((val) => val.pk);
    const liqorMangoAccounts = await Promise.all(liqors.map((liqor) => client.getMangoAccount(liqor, true)));
    liqorMangoAccounts.forEach((a) => {
        table.addRow({
            Account: a.publicKey.toBase58(),
            Equity: (0, utils_1.toUiDecimalsForQuote)(a.getEquity(group)).toLocaleString() + '$',
        });
    });
    const msg = title + '\n```\n' + table.render() + '\n```';
    console.log(msg);
    console.log();
}
async function main() {
    const client = await buildClient();
    const group = await client.getGroup(new web3_js_1.PublicKey(GROUP_PK));
    const mangoAccounts = await client.getAllMangoAccounts(group, true);
    const change = 0.4;
    const drop = 1 - change;
    const groupBear = (0, cloneDeep_1.default)(group);
    Array.from(groupBear.banksMapByTokenIndex.values())
        .flat()
        .forEach((b) => {
        b['oldUiPrice'] = b._uiPrice;
        b._uiPrice = b._uiPrice * drop;
        b._price = b._price?.mul(I80F48_1.I80F48.fromNumber(drop));
    });
    Array.from(groupBear.perpMarketsMapByMarketIndex.values()).forEach((p) => {
        p['oldUiPrice'] = p._uiPrice;
        p._uiPrice = p._uiPrice * drop;
        p._price = p._price?.mul(I80F48_1.I80F48.fromNumber(drop));
    });
    const rally = 1 + change;
    const groupBull = (0, cloneDeep_1.default)(group);
    Array.from(groupBull.banksMapByTokenIndex.values())
        .flat()
        .forEach((b) => {
        b['oldUiPrice'] = b._uiPrice;
        b._uiPrice = b._uiPrice * rally;
        b._price = b._price?.mul(I80F48_1.I80F48.fromNumber(rally));
    });
    Array.from(groupBull.perpMarketsMapByMarketIndex.values()).forEach((p) => {
        p['oldUiPrice'] = p._uiPrice;
        p._uiPrice = p._uiPrice * rally;
        p._price = p._price?.mul(I80F48_1.I80F48.fromNumber(rally));
    });
    const healthThresh = 0;
    let tableName = `Liqors acquire liabs and assets. The assets and liabs are sum of max assets and max 
  liabs for any token which would be liquidated to fix the health of a mango account. 
  This would be the slippage they would face on buying-liabs/offloading-assets tokens acquired from unhealth accounts after a`;
    await computePriceImpactForLiqor(groupBear, mangoAccounts, healthThresh, `Table 1a: ${tableName} 20% drop`);
    await computePriceImpactForLiqor(groupBull, mangoAccounts, healthThresh, `Table 1b: ${tableName} 20% rally`);
    tableName = 'Perp notional that liqor need to liquidate after a ';
    await computePerpPositionsToBeLiquidated(groupBear, mangoAccounts, healthThresh, `Table 2a: ${tableName} 20% drop`);
    await computePerpPositionsToBeLiquidated(groupBull, mangoAccounts, healthThresh, `Table 2b: ${tableName} 20% rally`);
    await logLiqorEquity(client, group, (await (await (0, node_fetch_1.default)(`https://api.mngo.cloud/data/v4/stats/liqors-over_period?over_period=1MONTH`)).json()).map((data) => new web3_js_1.PublicKey(data['liqor'])), `Table 3: Equity of known liqors from last month`);
    await logLiqorEquity(client, group, [
        new web3_js_1.PublicKey('CtHuPg2ctVVV7nqmvVEcMtcWyJAgtZw9YcNHFQidjPgF'),
        new web3_js_1.PublicKey('F1SZxEDxxCSLVjEBbMEjDYqajWRJQRCZBwPQnmcVvTLV'),
        new web3_js_1.PublicKey('BGYWnqfaauCeebFQXEfYuDCktiVG8pqpprrsD4qfqL53'),
        new web3_js_1.PublicKey('9XJt2tvSZghsMAhWto1VuPBrwXsiimPtsTR8XwGgDxK2'),
    ], `Table 4: Equity of known makers from last month`);
    // TODO warning when wrapper asset on chain price has too much difference to oracle
    // TODO warning when slippage is higher than liquidation fee
    // TODO warning when liqors equity is too low
    // TODO warning when mm equity is too low
    // TODO all awaits are linear, should be parallelised to speed up script
}
main();
